// ==============================================//
// Fichier					: gauge.js
// Version 					: 1.4
// Dernière modification 	: 31/12/2021
// ==============================================//

'use strict';

class Gauge {
  constructor(configuration) {
    // default configuration settings
    const config = {
      size : 200,
	  arcPadding : 15,
	  arcWidth : 20,
	  pointerWidth : 6,
	  pointerTailLength : 5,
	  pointerTail : 'angular',
	  labelInset : 10,
	  minAngle : -90,
	  maxAngle : 90,
      margin : 10,
      minValue : 0,
      maxValue : 10,
      majorTicks : 5,
      lowThreshhold : 3,
      highThreshhold : 7,
      scale: 'linear',
	  ticksLength: 'large',
      lowThreshholdColor : '#009900',
      defaultColor : '#ffe500',
      highThreshholdColor : '#cc0000',
      transitionMs : 1000,
      displayUnit: 'Value',
	  leftFootVal: '0 à 99%',
	  rightFootVal: '99 à 100%',
	  roundMes: 1,
	  unitMesPre: '',
	  unitMesSuf: '',
	  localeMes : 'fr-FR'	  
    };

    this.config = Object.assign(config, configuration);

    const pointerHeadLength = this._radius() - 2*this.config.labelInset;
    this.lineData = [
        [this.config.pointerWidth / 2, 0],
        [0, -pointerHeadLength],
        [-(this.config.pointerWidth / 2), 0],
        [0, this.config.pointerTailLength],
        [this.config.pointerWidth / 2, 0]
      ];
	this.roundData = 'M '+(this.config.pointerWidth / 2)+' 0 L 0 '+(-pointerHeadLength)+' L '+(-(this.config.pointerWidth / 2))+' 0 A '+(this.config.pointerWidth / 2)+' '+(this.config.pointerWidth / 2)+' 0 0 0 '+(this.config.pointerWidth / 2)+' 0';

    this.angleRange = this.config.maxAngle - this.config.minAngle;

    if (config.scale == 'log') {
      this.scale = d3.scaleLog()
        .range([0,1])
        .domain([config.minValue, config.maxValue]);
    }
    else {
      this.scale = d3.scaleLinear()
      .range([0,1])
      .domain([config.minValue, config.maxValue]);
    }

    const colorDomain = [config.lowThreshhold, config.highThreshhold].map(this.scale);
    const colorRange  = [
      config.lowThreshholdColor,
      config.defaultColor,
      config.highThreshholdColor
    ];
    this.colorScale = d3.scaleThreshold().domain(colorDomain).range(colorRange);

    let ticks = config.majorTicks;
    if (config.scale === 'log') {
      ticks = Math.log10(config.maxValue/minValue);
    }
    this.ticks = this.scale.ticks(ticks);

    this.threshholds = [
      config.minValue,
      config.lowThreshhold,
      config.highThreshhold,
      config.maxValue
    ]
    .map(d => this.scale(d));

    this.arc = d3.arc()
      .innerRadius(this._radius() - this.config.arcWidth - this.config.arcPadding)
      .outerRadius(this._radius() - this.config.arcPadding)
      .startAngle((d, i) => {
        const ratio = i > 0 ? this.threshholds[i-1] : this.threshholds[0];
        return this._deg2rad(this.config.minAngle + (ratio * this.angleRange));
      })
      .endAngle((d, i) => this._deg2rad(this.config.minAngle + this.threshholds[i] * this.angleRange));

  }

  _radius() {

    return (this.config.size - this.config.margin) / 2;

  }

  _deg2rad(deg) {

    return deg * Math.PI / 180;

  }

  render(container, newValue) {

  const svg = d3.select(container)
    .append('svg')
    .attr('class', 'gauge')
    .attr('width', this.config.size + this.config.margin)
    .attr('height', this.config.size / 2 + this.config.margin);

  // display panel arcs with color scale
  const arcs = svg.append('g')
    .attr('class', 'arc')
    .attr('transform', `translate(${this._radius()}, ${this._radius()})`);

  // draw the color arcs
  arcs.selectAll('path')
    .data(this.threshholds)
    .enter()
    .append('path')
    .attr('fill', d => this.colorScale(d-0.001))
    .attr('d', this.arc);

  // display panel - labels
  const lg = svg.append('g')
    .attr('class', 'label')
    .attr('transform', `translate(${this._radius()},${this._radius()})`);

  // display panel - text
  lg.selectAll('text')
    .data(this.ticks)
    .enter()
    .append('text')
    .attr('transform', d => {
       var newAngle = this.config.minAngle + (this.scale(d) * this.angleRange);
       return `rotate(${newAngle}) translate(0, ${this.config.labelInset - this._radius()})`;
     })
    .text(d3.format('1,.0f'));

  // display panel - ticks
  lg.selectAll('line')
    .data(this.ticks)
    .enter()
    .append('line')
    .attr('class', 'tickline')
    .attr('x1', 0)
    .attr('y1', 20-this.config.arcWidth)
    .attr('x2', 0);
   
   if(this.config.ticksLength == 'large'){
	lg.selectAll('line')
	  .attr('y2', 30);
   }
   else{
	lg.selectAll('line')
	  .attr('y2', 30-this.config.arcWidth);
   }
   
   lg.selectAll('line')
     .attr('transform', d => {
      const newAngle = this.config.minAngle + (this.scale(d) * this.angleRange);
      return `rotate(${newAngle}), translate(0, ${this.config.arcWidth  - this.config.labelInset - this._radius()})`;
    });

  // display pointer
  const pg = svg.append('g')
    .data([this.lineData])
	.attr('class', 'pointer')
    .attr('transform', `translate(${this._radius()},${this._radius()})`);

  const pointer = pg.append('path');
  if(this.config.pointerTail == 'angular'){
	pointer.attr('d', d3.line());
  }else{
    pointer.attr('d',this.roundData);
  }
  pointer.attr('transform', `rotate(${this.config.minAngle})`);
  
  // display arc feet value
  if((this.config.leftFootVal !== '') || (this.config.rightFootVal !== '')){
	  const FootDiv = d3.select(container).append('div')
		.style('width', `${(this.config.size - this.config.margin)}px`);

	  const leftFootVal = FootDiv.append('div')
		.attr('class', 'left-foot')
		.style('width', `${(this.config.size - this.config.margin)/2}px`)
		.text(d => this.config.leftFootVal)
		.style('color', this.config.lowThreshholdColor);

	  const rightFootVal = FootDiv.append('div')
		.attr('class', 'right-foot')
		.style('width', `${(this.config.size - this.config.margin)/2}px`)
		.text(d => this.config.rightFootVal)
		.style('color', this.config.highThreshholdColor);
  }
  
  // display current value
  const numberDiv = d3.select(container).append('div')
    .attr('class', 'number-div')
    .style('width', `${this.config.size - this.config.margin}px`);

  const numberUnit = numberDiv.append('span')
    .attr('class', 'number-unit')
    .text(d => this.config.displayUnit);

  const numberValue = numberDiv.append('span')
    .data([newValue])
    .attr('class', 'number-value')
    .text(d => d === undefined ? 0: d);
		
  this.pointer = pointer;
  this.numberValue = numberValue;

  }

  update(newValue) {
	
	var newAngle = this.config.minAngle;
	if(this.config.maxValue < newValue){
		newAngle = this.config.maxAngle;
	}else
	if(this.config.minValue < newValue){
		newAngle = this.config.minAngle + (this.scale(newValue) * this.angleRange);
	}
	
	const round = this.config.roundMes;
	const unit_pre = this.config.unitMesPre;
	const unit_suf = this.config.unitMesSuf;
	const locale_mes = this.config.localeMes;
	
    this.pointer.transition()
      .duration(this.config.transitionMs)
      .attr('transform', `rotate(${newAngle})`);

    this.numberValue
      .data([newValue])
      .transition()
      .duration(this.config.transitionMs)
      .style('color', this.colorScale( this.scale(newValue) ))
      .tween("", function(d) {
        const interpolator = d3.interpolate( this.textContent, d );
        const that = this;
        return function( t ) {
		  that.textContent = unit_pre+interpolator(t).toLocaleString(locale_mes,{minimumFractionDigits: round, maximumFractionDigits: round})+unit_suf;
        };
      });
  }

}